-- Checking traits to be valid will be postponed until a later version.

-----------------------------------------------------------------------
-- | Writing to the User Interface | ----------------------------------
-----------------------------------------------------------------------

-- printToFrame :: (Age a, Alignment a, Class a, Faith a, Feature a, Gender a, Name a, Race a, Tribe a, Motivations a) => a -> ()
local function printToFrame(characterTemplate)
  local tribeContext = getTribeContext(characterTemplate:getRace())
  frm_Main.charTribe:SetText(
    eitherOr(
      nilOrWhitespace(tribeContext),
      "",
      tribeContext))
  
  box_Age:SetText(        ucWords(                                             characterTemplate:getAge()        ))
  box_Alignment:SetText(  ucWords(                                             characterTemplate:getAlignment()  ))
  box_Class:SetText(      ucWords(                                             characterTemplate:getClass()      ))
  box_Faith:SetText(      ucWords(processSubstitutionString(characterTemplate, characterTemplate:getFaith()      )))
  box_Feature:SetText(    ucFirst(                                             characterTemplate:getFeature()    ))
  box_Gender:SetText(     ucWords(                                             characterTemplate:getGender()     ))
  box_Name:SetText(       ucWords(                                             characterTemplate:getName()       ))
  box_Race:SetText(       ucWords(                                             characterTemplate:getRace()       ))  
  box_Tribe:SetText(      ucWords(                                             characterTemplate:getTribe()      ))
  
  box_Motivation1:SetText(ucFirst(processSubstitutionString(characterTemplate, characterTemplate:getMotivation(1))))
  box_Motivation2:SetText(ucFirst(processSubstitutionString(characterTemplate, characterTemplate:getMotivation(2))))
  box_Motivation3:SetText(ucFirst(processSubstitutionString(characterTemplate, characterTemplate:getMotivation(3))))
end
  --[[
  characterTemplate.setMotivation = function(self, i, x) 
    x = processSubstitutionString(characterTemplate, x)
    if i == 1 then
      return box_Motivation1:SetText(ucFirst(x))
    elseif i == 2 then
      return box_Motivation2:SetText(ucFirst(x))
    elseif i == 3 then
      if box_Motivation2:GetText() == "" and x ~= "" then
        return box_Motivation2:SetText(ucFirst(x))
      else
        return box_Motivation3:SetText(ucFirst(x))
      end
    end
  end
  ]]

-- printToChat :: (Age a, Class a, Feature a, Gender a, Race a, Tribe a)
function printToChat()
  characterTemplate = readFromFrame(characterTemplate)
	local charString = generateCharacterString(characterTemplate)
  SendChatMessage(charString, "EMOTE")
end

-----------------------------------------------------------------------
-- | Reading from the User Interface | --------------------------------
-----------------------------------------------------------------------

function readAge()
  characterTemplate:setAge(box_Age:GetText():lower())
end

function readAlignment()
  characterTemplate:setAlignment(box_Alignment:GetText():lower())
end

function readClass()
  characterTemplate:setClass(box_Class:GetText():lower())
end

function readFaith()
  characterTemplate:setFaith(box_Faith:GetText():lower())
end

function readFeature()
  characterTemplate:setFeature(box_Feature:GetText())
end

function readGender()
  characterTemplate:setGender(box_Gender:GetText():lower())
end

function readName()
  characterTemplate:setName(box_Name:GetText())
end

function readRace()
  characterTemplate:setRace(box_Race:GetText():lower())
end

function readTribe()
  characterTemplate:setTribe(box_Tribe:GetText():lower())
end

function readMotivation(i)
  local boxes = { box_Motivation1, box_Motivation2, box_Motivation3 }
  characterTemplate:setMotivation(i, boxes[i]:GetText())
end

function readFromFrame(characterTemplate)
  --[[
  -- Problem, inverseXim and inverseXir both try to capture "her"
  -- Try alternative method: Capture the focus of the input fields, when they lose focus, write back to the characterTemplate.
  characterTemplate:setAge(                          box_Age:       GetText():lower())
  characterTemplate:setAlignment(                    box_Alignment: GetText():lower())
  characterTemplate:setClass(                        box_Class:     GetText():lower())
  characterTemplate:setFaith(     inverseFixGender(  box_Faith:     GetText():lower()))   -- Inverse SubSeq processing required 
  characterTemplate:setFeature(   inverseFixGender(  box_Feature:   GetText()        ))   -- Inverse SubSeq processing required
  characterTemplate:setGender(                       box_Gender:    GetText():lower())
  characterTemplate:setName(                         box_Name:      GetText():lower())
  characterTemplate:setRace(                         box_Race:      GetText():lower())
  characterTemplate:setTribe(                        box_Tribe:     GetText():lower())
  
  characterTemplate:setMotivation(1, inverseFixGender(  box_Motivation1:GetText())) -- Inverse SubSeq processing required 
  characterTemplate:setMotivation(2, inverseFixGender(  box_Motivation2:GetText())) -- Inverse SubSeq processing required 
  characterTemplate:setMotivation(3, inverseFixGender(  box_Motivation3:GetText())) -- Inverse SubSeq processing required 
  ]]
  return characterTemplate
end

-----------------------------------------------------------------------
-- | Rolling Traits | -------------------------------------------------
-----------------------------------------------------------------------

function rollName()
  characterTemplate = readFromFrame(characterTemplate)
  characterTemplate:setName(
    underscoreToSpace(
			doubleUnderscoreToApostrophe(
        generateRandomName(
          characterTemplate	) ) ) )
  printToFrame(characterTemplate)
end

function rollRace()
  characterTemplate = readFromFrame(characterTemplate)
  characterTemplate:setRace(
		generateRandomRace("any")
	)
	-- Also roll the Tribe to make sure it doesn't get invalid.
	rollTribe()
  printToFrame(characterTemplate)
end

function rollTribe()
  --characterTemplate = readFromFrame(characterTemplate)
	characterTemplate:setTribe(
		generateRandomTribe(
			characterTemplate
		)
	)
  printToFrame(characterTemplate)
end

function rollGender()
  characterTemplate = readFromFrame(characterTemplate)
  characterTemplate:setGender(
		generateRandomGender()
	)
  printToFrame(characterTemplate)
end

function rollClass()
  characterTemplate = readFromFrame(characterTemplate)
  characterTemplate:setClass(
    generateRandomClass(
      characterTemplate
    )
  )
  printToFrame(characterTemplate)
end

function rollAge()
  characterTemplate = readFromFrame(characterTemplate)
  characterTemplate:setAge(
    generateRandomAge(
      characterTemplate
    )
  )
  printToFrame(characterTemplate)
end

function rollAlignment()
  characterTemplate = readFromFrame(characterTemplate)
  characterTemplate:setAlignment(
    generateRandomAlignment(
			characterTemplate
    )
  )
  printToFrame(characterTemplate)
end

function rollFeature()
  characterTemplate = readFromFrame(characterTemplate)
  characterTemplate:setFeature(
    generateRandomFeature(
      characterTemplate
    )
  )
  printToFrame(characterTemplate)
end

function rollFaith()
  characterTemplate = readFromFrame(characterTemplate)
  characterTemplate:setFaith(
    generateRandomFaith(
      characterTemplate
    )
  )
  printToFrame(characterTemplate)
end

-- Function too complicated. Consider simplifying.
function rollMotivation(num)
  characterTemplate = readFromFrame(characterTemplate)
  local motivation = generateRandomMotivation(
    characterTemplate)
  
  -- If no options can be found, it might be impossible to find unique traits.
  -- Is this even still relevant?
	-- We can use nil-monadic operations to do this?
  if motivation == "Error: No options found." then
    characterTemplate:setMotivation(num, motivation)
  else
  
  local a, b = unpack(filter(function(x) return x ~= num end, { 1,2,3 }))
  
    -- Function 2
    while 
      motivation == characterTemplate:getMotivation(a) or 
      motivation == characterTemplate:getMotivation(b)
		do
			motivation = generateRandomMotivation(
				characterTemplate)
    end
    -- End of Function 2
    
    characterTemplate:setMotivation(num, motivation)
  end
  printToFrame(characterTemplate)
end

function rollAllMotivations()
  local motivations = generateRandomMotivations(
    characterTemplate
	)
  characterTemplate:setMotivation(1, motivations[1] or "Error: No options found.")
  characterTemplate:setMotivation(2, motivations[2] or "")  -- Finding less than three motivations is fine, as long as at least one motivation is found.
  characterTemplate:setMotivation(3, motivations[3] or "")
  printToFrame(characterTemplate)
end

local inc = getIncremental(1)
local activeModules = { }
activeModules[inc()] = { f =           rollRace, enabled = true }
activeModules[inc()] = { f =          rollTribe, enabled = true }
activeModules[inc()] = { f =            rollAge, enabled = true }
activeModules[inc()] = { f =         rollGender, enabled = true }
activeModules[inc()] = { f =      rollAlignment, enabled = true }
activeModules[inc()] = { f =          rollFaith, enabled = true }
activeModules[inc()] = { f =          rollClass, enabled = true }
activeModules[inc()] = { f =           rollName, enabled = true }
activeModules[inc()] = { f =        rollFeature, enabled = true }
activeModules[inc()] = { f = rollAllMotivations, enabled = true }

function rollAll(self)
  characterTemplate = nil
  characterTemplate = CharacterTemplate:new()
  invokeAll(
    map(
      function(x) return x.f end,
      filter(
        function(y) return y.enabled == true end,
        activeModules
      )
    )
  )
  printToFrame(characterTemplate)
end

-----------------------------------------------------------------------
-- | Locking Traits | -------------------------------------------------
-----------------------------------------------------------------------

function lockAge()
  activeModules[
    findKey(
      function(x) return x.f == rollAge end,
      activeModules
    )
  ].enabled = frm_Main.lockAge:GetChecked()
end

function lockAlignment()
  activeModules[
    findKey(
      function(x) return x.f == rollAlignment end,
      activeModules
    )
  ].enabled = frm_Main.lockAlignment:GetChecked()
end

function lockClass()
  activeModules[
    findKey(
      function(x) return x.f == rollClass end,
      activeModules
    )
  ].enabled = frm_Main.lockClass:GetChecked()
end

function lockFaith()
  activeModules[
    findKey(
      function(x) return x.f == rollFaith end,
      activeModules
    )
  ].enabled = frm_Main.lockFaith:GetChecked()
end

function lockFeature()
  activeModules[
    findKey(
      function(x) return x.f == rollFeature end,
      activeModules
    )
  ].enabled = frm_Main.lockFeature:GetChecked()
end

function lockGender()
  activeModules[
    findKey(
      function(x) return x.f == rollGender end,
      activeModules
    )
  ].enabled = frm_Main.lockGender:GetChecked()
end

function lockName()
  activeModules[
    findKey(
      function(x) return x.f == rollName end,
      activeModules
    )
  ].enabled = frm_Main.lockName:GetChecked()
end

function lockRace()
  activeModules[
    findKey(
      function(x) return x.f == rollRace end,
      activeModules
    )
  ].enabled = frm_Main.lockRace:GetChecked()
end

function lockTribe()
  activeModules[
    findKey(
      function(x) return x.f == rollTribe end,
      activeModules
    )
  ].enabled = frm_Main.lockTribe:GetChecked()
end

function lockMotivation(i)
  print "Error: Locking motivations not supported yet."
end

-- End of File

-----------------------------------------------------------------------
-- | Clearing Traits | ------------------------------------------------
-----------------------------------------------------------------------

function clearName()
  characterTemplate = readFromFrame(characterTemplate)
  characterTemplate:setName("")
  --frm_Main.lockName:SetChecked("true")
  printToFrame(characterTemplate)
end

function clearAge()
  characterTemplate = readFromFrame(characterTemplate)
  characterTemplate:setAge("")
  --frm_Main.lockAge:SetChecked("true")
  printToFrame(characterTemplate)
end

function clearRace()
  characterTemplate = readFromFrame(characterTemplate)
  characterTemplate:setRace("")
  --frm_Main.lockRace:SetChecked("true")
  printToFrame(characterTemplate)
end

function clearClass()
  characterTemplate = readFromFrame(characterTemplate)
  characterTemplate:setClass("")
  --frm_Main.lockClass:SetChecked("true")
  printToFrame(characterTemplate)
end

function clearGender()
  characterTemplate = readFromFrame(characterTemplate)
  characterTemplate:setGender("")
  --frm_Main.lockGender:SetChecked("true")
  printToFrame(characterTemplate)
end

function clearAlignment()
  characterTemplate = readFromFrame(characterTemplate)
  characterTemplate:setAlignment("")
  --frm_Main.lockAlignment:SetChecked("true")
  printToFrame(characterTemplate)
end

function clearFaith()
  characterTemplate = readFromFrame(characterTemplate)
  characterTemplate:setFaith("")
  --frm_Main.lockFaith:SetChecked("true")
  printToFrame(characterTemplate)
end

function clearTribe()
  characterTemplate = readFromFrame(characterTemplate)
  characterTemplate:setTribe("")
  --frm_Main.lockTribe:SetChecked("true")
  printToFrame(characterTemplate)
end

function clearFeature()
  characterTemplate = readFromFrame(characterTemplate)
  characterTemplate:setFeature("")
  --frm_Main.lockFeature:SetChecked("true")
    printToFrame(characterTemplate)
end

function clearMotivation(i)
  characterTemplate = readFromFrame(characterTemplate)
  if i == 1 then
    characterTemplate:setMotivation(1, box_Motivation2:GetText())
    characterTemplate:setMotivation(2, box_Motivation3:GetText())
  elseif i == 2 then
    characterTemplate:setMotivation(2, box_Motivation3:GetText())
  end
  
  characterTemplate:setMotivation(3, "")
  
  printToFrame(characterTemplate)
end

function clearAll()
  clearAge()
  clearAlignment()
  clearClass()
  clearFaith()
  clearFeature()
  clearGender()
  clearName()
  clearRace()
  clearTribe()
  clearMotivation(3)
  clearMotivation(2)
  clearMotivation(1)
  frm_Main.charTribe:SetText("Tribe")
  printToFrame(characterTemplate)
end